# coding: utf-8
"""

"""
import hashlib
import os
import secrets
import typing

import itsdangerous
import flask
import flask_login
import requests.exceptions

from flask_babel import _

from . import frontend
from .federation_forms import AddComponentForm, EditComponentForm, SyncComponentForm, CreateAPITokenForm, \
    AddOwnAPITokenForm, AuthenticationMethodForm, AddAliasForm, EditAliasForm, DeleteAliasForm, ModifyELNIdentityForm, \
    FederatedUserCreationForm
from ..logic import errors
from ..logic.utils import send_federated_login_account_creation_email
from .utils import check_current_user_is_not_readonly
from ..logic.component_authentication import remove_component_authentication_method, add_token_authentication, remove_own_component_authentication_method, add_own_token_authentication
from ..logic.components import get_component, update_component, add_component, get_components, check_component_exists, get_component_infos, ComponentInfo, get_component_by_uuid
from ..logic.federation.update import import_updates
from ..logic.federation.utils import login_required_saml, federated_login_route
from ..logic.federation.login import sp_login, get_idp_metadata, get_sp_metadata, process_login, resolve_artifact, eval_artifact_resolve, delete_old_artifacts
from ..logic.users import create_user, get_user_aliases_for_user, create_user_alias, update_user_alias, delete_user_alias, get_user_alias, delete_user_link, \
    get_federated_identity, get_federated_identities, revoke_user_links_by_fed_ids, create_sampledb_federated_identity, delete_user_links_by_fed_ids, \
    revoke_user_link, enable_user_link, get_user_by_federated_user, get_user, get_active_federated_identity_by_federated_user, enable_federated_identity_for_login, identity_not_required_for_auth, set_user_hidden
from ..logic.background_tasks import post_poke_components_task
from ..models import OwnComponentAuthentication, ComponentAuthenticationType, ComponentAuthentication, User, UserType, BackgroundTaskStatus, Authentication, AuthenticationType
from ..utils import FlaskResponseT


FEDERATED_IDENTITY_TOKENS_MAX_AGE = 60 * 5  # seconds


@frontend.route('/other-databases/<int:component_id>', methods=['GET', 'POST'])
@flask_login.login_required
def component(component_id: int) -> FlaskResponseT:
    try:
        component = get_component(component_id)
    except errors.ComponentDoesNotExistError:
        return flask.abort(404)
    try:
        alias = get_user_alias(flask_login.current_user.get_id(), component_id)
    except errors.UserAliasDoesNotExistError:
        alias = None
    created_api_token = None
    add_own_api_token_form = AddOwnAPITokenForm()
    create_api_token_form = CreateAPITokenForm()
    authentication_method_form = AuthenticationMethodForm()
    own_authentication_method_form = AuthenticationMethodForm()
    edit_component_form = EditComponentForm()
    show_edit_form = False
    if edit_component_form.address.data is None:
        if component.address is None:
            edit_component_form.address.data = ''
        else:
            edit_component_form.address.data = component.address
    if edit_component_form.name.data is None:
        if component.name is None:
            edit_component_form.name.data = ''
        else:
            edit_component_form.name.data = component.name
    if edit_component_form.description.data is None:
        edit_component_form.description.data = component.description
    sync_component_form = SyncComponentForm()

    own_api_tokens = OwnComponentAuthentication.query.filter(OwnComponentAuthentication.component_id == component_id, OwnComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()
    api_tokens = ComponentAuthentication.query.filter(ComponentAuthentication.component_id == component_id, ComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()

    if 'edit' in flask.request.form:
        show_edit_form = True
        if edit_component_form.validate_on_submit():
            check_current_user_is_not_readonly()
            if not flask_login.current_user.is_admin:
                return flask.abort(403)
            name = edit_component_form.name.data
            address = edit_component_form.address.data
            if address == '':
                address = None
            if name == '':
                name = None
            try:
                update_component(component_id=component_id, name=name, description=edit_component_form.description.data, address=address)
            except errors.ComponentDoesNotExistError:
                edit_component_form.name.errors.append(_('This database does not exist'))
            except errors.InvalidComponentNameError:
                edit_component_form.name.errors.append(_('This database name is invalid'))
            except errors.InvalidComponentAddressError:
                edit_component_form.address.errors.append(_('This database address is invalid'))
            except errors.ComponentAlreadyExistsError:
                edit_component_form.name.errors.append(_('A database with this UUID or name has already been added'))
                edit_component_form.address.errors.append(_('A database with this UUID or name has already been added'))
            except errors.InvalidComponentUUIDError:
                edit_component_form.name.errors.append(_('Invalid UUID'))
            except errors.InsecureComponentAddressError:
                edit_component_form.address.errors.append(_('Only secure communication via https is allowed'))
            except Exception:
                edit_component_form.name.errors.append(_('Failed to update database'))
            else:
                post_poke_components_task()
                flask.flash(_('Database information updated successfully'), 'success')
                return flask.redirect(flask.url_for('.component', component_id=component_id))
    else:
        if component.address is None:
            edit_component_form.address.data = ''
        else:
            edit_component_form.address.data = component.address
        if component.name is None:
            edit_component_form.name.data = ''
        else:
            edit_component_form.name.data = component.name
        edit_component_form.description.data = component.description
    if 'sync' in flask.request.form and sync_component_form.validate_on_submit():
        check_current_user_is_not_readonly()
        if not flask_login.current_user.is_admin:
            return flask.abort(403)
        ignore_last_sync_time = sync_component_form.ignore_last_sync_time.data
        try:
            import_updates(component, ignore_last_sync_time=ignore_last_sync_time)
            flask.flash(_('Successfully imported data changes.'), 'success')
        except errors.MissingComponentAddressError:
            flask.flash(_('Missing database address.'), 'error')
        except errors.NoAuthenticationMethodError:
            flask.flash(_('No valid authentication method configured.'), 'error')
        except errors.UnauthorizedRequestError:
            flask.flash(_('Invalid authentication method.'), 'error')
        except errors.InvalidDataExportError as error:
            flask.flash(_('Received invalid data. Error message: "%(error)s"', error=str(error)), 'error')
        except errors.ComponentNotConfiguredForFederationError:
            flask.flash(_('This database has not been configured to exchange data with other databases.'))
        except errors.RequestServerError:
            flask.flash(_('Server error when requesting the data.'), 'error')
        except errors.RequestError:
            flask.flash(_('Received error code when requesting the data.'), 'error')
        except requests.exceptions.ConnectionError:
            flask.flash(_('Failed to establish a connection to %(component_name)s (%(component_address)s).', component_name=component.get_name(), component_address=component.address), 'error')
        return flask.redirect(flask.url_for('.component', component_id=component_id))
    if 'remove' in flask.request.form and flask.request.form['remove'] == 'Remove':
        check_current_user_is_not_readonly()
        if not flask_login.current_user.is_admin:
            return flask.abort(403)
        authentication_method_id = authentication_method_form.id.data
        if authentication_method_form.validate_on_submit():
            try:
                remove_component_authentication_method(authentication_method_id)
                flask.flash(_('Successfully removed the authentication method.'), 'success')
                api_tokens = ComponentAuthentication.query.filter(ComponentAuthentication.component_id == component_id, ComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()
            except errors.AuthenticationMethodDoesNotExistError:
                flask.flash(_('Authentication method has already been deleted.'), 'error')
            except Exception:
                flask.flash(_('Failed to remove the authentication method.'), 'error')
    if 'removeOwn' in flask.request.form and flask.request.form['removeOwn'] == 'RemoveOwn':
        check_current_user_is_not_readonly()
        if not flask_login.current_user.is_admin:
            return flask.abort(403)
        authentication_method_id = authentication_method_form.id.data
        if authentication_method_form.validate_on_submit():
            try:
                remove_own_component_authentication_method(authentication_method_id)
                flask.flash(_('Successfully removed the authentication method.'), 'success')
                own_api_tokens = OwnComponentAuthentication.query.filter(OwnComponentAuthentication.component_id == component_id, OwnComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()
            except errors.AuthenticationMethodDoesNotExistError:
                flask.flash(_('Authentication method has already been deleted.'), 'error')
            except Exception:
                flask.flash(_('Failed to remove the authentication method.'), 'error')
    if 'create_api_token' in flask.request.form and create_api_token_form.validate_on_submit():
        check_current_user_is_not_readonly()
        if not flask_login.current_user.is_admin:
            return flask.abort(403)
        created_api_token = secrets.token_hex(32)
        description = create_api_token_form.description.data
        try:
            add_token_authentication(component_id, created_api_token, description)
        except Exception:
            flask.flash(_('Failed to add API token.'), 'error')
        api_tokens = ComponentAuthentication.query.filter(ComponentAuthentication.component_id == component_id, ComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()
    if 'add_own_api_token' in flask.request.form and add_own_api_token_form.validate_on_submit():
        check_current_user_is_not_readonly()
        if not flask_login.current_user.is_admin:
            return flask.abort(403)
        description = add_own_api_token_form.description.data
        try:
            add_own_token_authentication(component_id, add_own_api_token_form.token.data, description)
            own_api_tokens = OwnComponentAuthentication.query.filter(OwnComponentAuthentication.component_id == component_id, OwnComponentAuthentication.type == ComponentAuthenticationType.TOKEN).all()
        except errors.InvalidTokenError:
            flask.flash(_('Invalid token. Required length: 64 digits.'), 'error')
        except errors.TokenExistsError:
            flask.flash(_('This token has already been linked to this database.'), 'error')
        except Exception:
            flask.flash(_('Failed to add API token.'), 'error')

    active_identities = get_federated_identities(flask_login.current_user.id, component, active_status=True)
    inactive_identities = get_federated_identities(flask_login.current_user.id, component, active_status=False)

    component_address = component.address
    if component_address and not component_address.endswith('/'):
        component_address += '/'

    has_other_auth_methods = identity_not_required_for_auth(component_id)

    return flask.render_template(
        'other_databases/component.html',
        component=component,
        alias=alias,
        show_edit_form=show_edit_form,
        edit_component_form=edit_component_form,
        sync_component_form=sync_component_form,
        add_own_api_token_form=add_own_api_token_form,
        create_api_token_form=create_api_token_form,
        authentication_method_form=authentication_method_form,
        own_authentication_method_form=own_authentication_method_form,
        api_tokens=api_tokens,
        own_api_tokens=own_api_tokens,
        created_api_token=created_api_token,
        active_identities=active_identities,
        inactive_identities=inactive_identities,
        federation_user_base_url=f"{component_address}users",
        has_other_auth_methods=has_other_auth_methods
    )


@frontend.route('/other-databases/', methods=['GET', 'POST'])
@flask_login.login_required
def federation() -> FlaskResponseT:
    components = get_components()

    # federation graph
    nodes = []
    edges = []
    if flask.current_app.config['FEDERATION_UUID']:
        node = {
            'label': flask.current_app.config['SERVICE_NAME'],
            'title': flask.current_app.config['FEDERATION_UUID'],
            'borderWidth': 2
        }
        nodes_by_uuid = {
            flask.current_app.config['FEDERATION_UUID']: node
        }
        reachable_uuids = {
            flask.current_app.config['FEDERATION_UUID']
        }

        for component in components:
            if component.uuid == flask.current_app.config['FEDERATION_UUID'] or not component.discoverable:
                continue
            node = {
                'label': component.get_name(),
                'title': component.uuid
            }
            nodes_by_uuid[component.uuid] = node
            reachable_uuids.add(component.uuid)

        component_infos_by_uuid: typing.Dict[str, typing.List[ComponentInfo]] = {}
        all_component_infos = get_component_infos()
        for component_info in all_component_infos:
            if not component_info.discoverable:
                continue
            if component_info.uuid not in component_infos_by_uuid:
                component_infos_by_uuid[component_info.uuid] = []
            component_infos_by_uuid[component_info.uuid].append(component_info)
        did_update = True
        while did_update:
            did_update = False
            for uuid, component_infos in component_infos_by_uuid.items():
                if uuid in reachable_uuids:
                    continue
                for component_info in component_infos:
                    if component_info.discoverable and component_info.source_uuid in reachable_uuids:
                        reachable_uuids.add(component_info.uuid)
                        did_update = True
        for uuid, component_infos in component_infos_by_uuid.items():
            if uuid in nodes_by_uuid or uuid not in reachable_uuids:
                continue
            component_infos_by_distance: typing.Dict[int, typing.List[ComponentInfo]] = {}
            for component_info in component_infos:
                if component_info.distance not in component_infos_by_distance:
                    component_infos_by_distance[component_info.distance] = []
                component_infos_by_distance[component_info.distance].append(component_info)
            component_infos = component_infos_by_distance[min(component_infos_by_distance)]
            names = {
                component_info.name
                for component_info in component_infos
                if component_info.name
            }
            if not names:
                name = None
            elif len(names) == 1:
                name = list(names)[0]
            else:
                name = ' / '.join(sorted(names))
            node = {
                'label': name or uuid,
                'title': uuid
            }
            nodes_by_uuid[uuid] = node

        node_ids_by_uuid = {
            uuid: node_id
            for node_id, uuid in enumerate(nodes_by_uuid)
        }
        for uuid, node in nodes_by_uuid.items():
            node['id'] = node_ids_by_uuid[uuid]
        nodes = list(nodes_by_uuid.values())

        for component in components:
            if component.uuid in node_ids_by_uuid and component.discoverable:
                edges.append({
                    'from': node_ids_by_uuid[flask.current_app.config['FEDERATION_UUID']],
                    'to': node_ids_by_uuid[component.uuid],
                    'arrows': {
                        (True, True): 'from, to',
                        (True, False): 'from',
                        (False, True): 'to',
                        (False, False): None
                    }.get((component.import_token_available, component.export_token_available))
                })
        for component_info in all_component_infos:
            if component_info.source_uuid in node_ids_by_uuid and component_info.uuid in node_ids_by_uuid and component_info.discoverable:
                edges.append({
                    'from': node_ids_by_uuid[component_info.source_uuid],
                    'to': node_ids_by_uuid[component_info.uuid],
                    'dashes': True
                })

    add_component_form = AddComponentForm()
    if add_component_form.address.data is None:
        add_component_form.address.data = ''
    if add_component_form.uuid.data is None:
        add_component_form.uuid.data = ''
    if add_component_form.name.data is None:
        add_component_form.name.data = ''
    if add_component_form.description.data is None:
        add_component_form.description.data = ''
    show_add_form = False
    if 'add' in flask.request.form:
        show_add_form = True
        if add_component_form.validate_on_submit():
            check_current_user_is_not_readonly()
            try:
                name = add_component_form.name.data
                address = add_component_form.address.data
                if not flask.current_app.config['ALLOW_HTTP'] and address[:7] == 'http://':
                    add_component_form.address.errors.append(_('Only secure communication via https is allowed'))
                    return flask.render_template("other_databases/federation.html", current_user=flask_login.current_user, components=components, add_component_form=add_component_form, show_add_form=show_add_form)
                if name == '':
                    name = None
                if address == '':
                    address = None
                component_id = add_component(uuid=add_component_form.uuid.data, name=name, description=add_component_form.description.data, address=address).id
            except errors.ComponentAlreadyExistsError as err:
                if err.is_current_app:
                    add_component_form.name.errors.append(_('This UUID or name is already used by this instance (%(service_name)s)', service_name=flask.current_app.config['SERVICE_NAME']))
                    add_component_form.uuid.errors.append(_('This UUID or name is already used by this instance (%(service_name)s)', service_name=flask.current_app.config['SERVICE_NAME']))
                else:
                    add_component_form.name.errors.append(_('A database with this UUID or name has already been added'))
                    add_component_form.uuid.errors.append(_('A database with this UUID or name has already been added'))
            except errors.InvalidComponentNameError:
                add_component_form.name.errors.append(_('This database name is invalid'))
            except errors.InvalidComponentUUIDError:
                add_component_form.uuid.errors.append(_('Invalid UUID'))
            except errors.InsecureComponentAddressError:
                add_component_form.address.errors.append(_('Only secure communication via https is allowed'))
            except errors.InvalidComponentAddressError:
                add_component_form.address.errors.append(_('This database address is invalid'))
            except Exception:
                add_component_form.name.errors.append(_('Failed to add database'))
            else:
                post_poke_components_task()
                flask.flash(_('The database information has been added successfully'), 'success')
                return flask.redirect(flask.url_for('.component', component_id=component_id))
    return flask.render_template(
        "other_databases/federation.html",
        current_user=flask_login.current_user,
        components=components,
        add_component_form=add_component_form,
        show_add_form=show_add_form,
        nodes=nodes,
        edges=edges,
    )


@frontend.route('/other-databases/alias/', methods=['GET', 'POST'])
@flask_login.login_required
def user_alias() -> FlaskResponseT:
    user = flask_login.current_user
    components = get_components()
    aliases = get_user_aliases_for_user(user.id)
    added_components = [alias.component_id for alias in aliases]
    addable_components = [comp for comp in components if comp.id not in added_components]
    try:
        add_alias_component = int(flask.request.args.get('add_alias_component', ''))
        check_component_exists(add_alias_component)
    except ValueError:
        add_alias_component = None
    except TypeError:
        add_alias_component = None
    except errors.ComponentDoesNotExistError:
        add_alias_component = None
    add_alias_form = AddAliasForm()
    add_alias_form.component.choices = [
        (str(comp.id), comp.name)
        for comp in components if comp.id not in added_components
    ]

    edit_alias_form = EditAliasForm()

    delete_alias_form = DeleteAliasForm()

    show_edit_form = False

    if 'edit' in flask.request.form:
        show_edit_form = True
        if edit_alias_form.validate_on_submit():
            try:
                update_user_alias(
                    user.id, edit_alias_form.component.data,
                    edit_alias_form.name.data if edit_alias_form.name.data != '' and not edit_alias_form.use_real_name.data else None,
                    edit_alias_form.use_real_name.data,
                    None,
                    edit_alias_form.use_real_email.data,
                    None,
                    edit_alias_form.use_real_orcid.data,
                    edit_alias_form.affiliation.data if edit_alias_form.affiliation.data != '' and not edit_alias_form.use_real_affiliation.data else None,
                    edit_alias_form.use_real_affiliation.data,
                    edit_alias_form.role.data if edit_alias_form.role.data != '' and not edit_alias_form.use_real_role.data else None,
                    edit_alias_form.use_real_role.data,
                )
            except errors.ComponentDoesNotExistError:
                flask.flash(_('That database does not exist.'), 'error')
            except errors.UserAliasDoesNotExistError:
                flask.flash(_('There is no alias for this database.'), 'error')
            else:
                flask.flash(_('User alias updated successfully.'), 'success')
                return flask.redirect(flask.url_for('.user_alias'))
    if 'add' in flask.request.form:
        if add_alias_form.validate_on_submit():
            create_user_alias(
                user.id, add_alias_form.component.data,
                add_alias_form.name.data if add_alias_form.name.data != '' and not add_alias_form.use_real_name.data else None,
                add_alias_form.use_real_name.data,
                None,
                add_alias_form.use_real_email.data,
                None,
                add_alias_form.use_real_orcid.data,
                add_alias_form.affiliation.data if add_alias_form.affiliation.data != '' and not add_alias_form.use_real_affiliation.data else None,
                add_alias_form.use_real_affiliation.data,
                add_alias_form.role.data if add_alias_form.role.data != '' and not add_alias_form.use_real_role.data else None,
                add_alias_form.use_real_role.data,
            )
            flask.flash(_('User alias updated successfully.'), 'success')
            return flask.redirect(flask.url_for('.user_alias'))
    if 'delete' in flask.request.form:
        if delete_alias_form.validate_on_submit():
            try:
                delete_user_alias(flask_login.current_user.id, edit_alias_form.component.data)
            except errors.ComponentDoesNotExistError:
                flask.flash(_('That database does not exist.'), 'error')
            except errors.UserAliasDoesNotExistError:
                flask.flash(_('The alias for this database has already been deleted.'), 'error')
            else:
                flask.flash(_('User alias updated successfully.'), 'success')
                return flask.redirect(flask.url_for('.user_alias'))

    if add_alias_form.name.data is None:
        add_alias_form.name.data = user.name
    if add_alias_form.affiliation.data is None:
        add_alias_form.affiliation.data = user.affiliation
    if add_alias_form.role.data is None:
        add_alias_form.role.data = user.role
    if add_alias_component and not add_alias_form.is_submitted():
        add_alias_form.component.data = add_alias_component

    aliases_by_component = {
        alias.component_id: {
            'is_default': alias.is_default,
            'name': alias.name if alias.name is not None else '',
            'email': alias.email if alias.email is not None else '',
            'orcid': alias.orcid if alias.orcid is not None else '',
            'affiliation': alias.affiliation if alias.affiliation is not None else '',
            'role': alias.role if alias.role is not None else '',
            'use_real_name': alias.use_real_name,
            'use_real_email': alias.use_real_email,
            'use_real_orcid': alias.use_real_orcid,
            'use_real_affiliation': alias.use_real_affiliation,
            'use_real_role': alias.use_real_role,
        } for alias in aliases}

    user_data = {
        'name': user.name if user.name is not None else '',
        'email': user.email if user.email is not None else '',
        'orcid': user.orcid if user.orcid is not None else '',
        'affiliation': user.affiliation if user.affiliation is not None else '',
        'role': user.role if user.role is not None else ''
    }

    component_names = {component.id: component.get_name() for component in components}

    return flask.render_template(
        "other_databases/federation_alias.html",
        user=user,
        addable_components=addable_components,
        aliases=aliases,
        add_alias_form=add_alias_form if addable_components else None,
        edit_alias_form=edit_alias_form,
        delete_alias_form=delete_alias_form,
        aliases_by_component=aliases_by_component,
        user_data=user_data,
        component_names=component_names,
        show_edit_form=show_edit_form
    )


@frontend.route("/other-databases/redirect-identity-confirmation/<string:component>", methods=["POST"])
@flask_login.login_required
def redirect_login_confirmation(component: str) -> FlaskResponseT:
    try:
        database = get_component_by_uuid(component)
    except errors.InvalidComponentUUIDError:
        return flask.redirect(flask.url_for("frontend.federation"))
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='federated-identities')
    session_state = hashlib.sha256(os.urandom(1024)).hexdigest()
    flask.session['fim_state'] = session_state
    link_identity_token = serializer.dumps({'user_id': flask_login.current_user.id, 'state': session_state})

    component_address = database.address
    if component_address and not component_address.endswith('/'):
        component_address += '/'

    all_fed_identities = get_federated_identities(user_id=flask_login.current_user.id, component=database)
    all_component_fed_ids = {identity.local_fed_user.fed_id for identity in get_federated_identities(component=database)}
    different_linked_ids = all_component_fed_ids.difference(identity.local_fed_user.fed_id for identity in all_fed_identities)

    url_enc_verified_fed_identities = ",".join(str(identity.local_fed_user.fed_id) for identity in all_fed_identities if identity.login)
    url_enc_unverified_fed_identities = ",".join(str(identity.local_fed_user.fed_id) for identity in all_fed_identities if not identity.login)
    url_enc_different_linked_fed_identities = ",".join(str(id) for id in different_linked_ids)

    redirect_url = f"{component_address}other-databases/link-identity?source_db={flask.current_app.config['FEDERATION_UUID']}&token={str(link_identity_token)}&state={session_state}"

    if url_enc_verified_fed_identities:
        redirect_url += f"&verified_ids={url_enc_verified_fed_identities}"
    if url_enc_unverified_fed_identities:
        redirect_url += f"&unverified_ids={url_enc_unverified_fed_identities}"
    if url_enc_different_linked_fed_identities:
        redirect_url += f"&linked_ids={url_enc_different_linked_fed_identities}"

    return flask.redirect(redirect_url)


@frontend.route("/other-databases/finalize-link-identity/", methods=["GET"])
def link_identity() -> FlaskResponseT:
    identity_token = flask.request.args.get('token')
    federation_partner_uuid = flask.request.args.get('federation_partner_uuid')

    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='federated-identities')

    if not federation_partner_uuid:
        flask.flash(_("An error occurred: Invalid database UUID."), 'error')
        return flask.redirect(flask.url_for(".federation"))

    try:
        federation_partner = get_component_by_uuid(federation_partner_uuid)
    except errors.InvalidComponentUUIDError:
        flask.flash(_("Unknown database (%(uuid)s)", uuid=federation_partner_uuid), 'error')
        return flask.redirect(flask.url_for('.federation'))

    federation_partner_address = federation_partner.address
    if federation_partner_address and not federation_partner_address.endswith('/'):
        federation_partner_address += '/'
    result = requests.get(f"{federation_partner_address}other-databases/validate-identity-token/?token={identity_token}", timeout=60)
    if result.status_code != 200:
        return flask.redirect(flask.url_for(".federation"))

    result_json = result.json()
    token = result_json.get('token')
    fed_user_id = result_json.get('fed_user')

    try:
        token_data = serializer.loads(token, max_age=FEDERATED_IDENTITY_TOKENS_MAX_AGE)
    except itsdangerous.BadSignature:
        flask.flash(_("An error occurred: Users could not be linked."), 'error')
        return flask.redirect(flask.url_for(".federation"))

    if 'fim_state' not in flask.session or flask.session['fim_state'] != token_data['state']:
        flask.flash(_("An error occurred: Users could not be linked."), 'error')
        return flask.redirect(flask.url_for(".federation"))

    try:
        federation_user = get_user(user_id=fed_user_id, component_id=federation_partner.id)
        identity = get_federated_identity(token_data['user_id'], federation_user.id)
    except (errors.FederatedIdentityNotFoundError, errors.UserDoesNotExistError):
        identity = None

    try:
        if identity is None or not identity.active:
            create_sampledb_federated_identity(token_data['user_id'], federation_partner, fed_user_id, update_if_inactive=True, use_for_login=True)
            flask.flash(_("You were successfully linked with a federated user from %(component_name)s.", component_name=federation_partner.get_name()), 'success')
        else:
            enable_federated_identity_for_login(identity)
            flask.flash(_("The identity was successfully verified."), 'success')
    except errors.FederatedUserInFederatedIdentityError:
        flask.flash(_("The federated user already has a federated identity in this database."), 'error')

    return flask.redirect(flask.url_for('.component', component_id=federation_partner.id))


@frontend.route("/other-databases/redirect-uuid/<string:component>", methods=["GET"])
def redirect_by_federation_uuid(component: str) -> FlaskResponseT:
    try:
        database = get_component_by_uuid(component_uuid=component)
    except errors.ComponentDoesNotExistError:
        return flask.redirect(flask.url_for('.federation'))
    return flask.redirect(flask.url_for('.component', component_id=database.id))


@frontend.route("/other-databases/link-identity/", methods=["GET"])
@flask_login.login_required
def confirm_identity() -> FlaskResponseT:
    source_database_uuid = flask.request.args.get('source_db', '')
    token = flask.request.args.get('token')

    def int_or_none(s: str) -> typing.Optional[int]:
        try:
            return int(s)
        except ValueError:
            return None

    form_verified_ids = flask.request.args.get('verified_ids')
    form_unverified_ids = flask.request.args.get('unverified_ids')
    form_linked_ids = flask.request.args.get('linked_ids')
    verified_ids = [int_or_none(id) for id in form_verified_ids.split(',')] if form_verified_ids else []
    unverified_ids = [int_or_none(id) for id in form_unverified_ids.split(',')] if form_unverified_ids else []
    linked_ids = [int_or_none(id) for id in form_linked_ids.split(',')] if form_linked_ids else []
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='identity-token-validation')

    source_database = None

    try:
        source_database = get_component_by_uuid(source_database_uuid)
    except (errors.InvalidComponentUUIDError, errors.ComponentDoesNotExistError):
        flask.flash(_('Could not confirm identity: The databases are not in a federation.'), 'error')
        return flask.redirect(flask.url_for('.index'))

    identity_token = serializer.dumps({'fed_user': flask_login.current_user.id, 'token': token})
    component_address = source_database.address
    if component_address and not component_address.endswith('/'):
        component_address += '/'

    confirmation_url = f"{component_address}other-databases/finalize-link-identity"
    cancel_url = f"{component_address}other-databases/redirect-uuid/{flask.current_app.config['FEDERATION_UUID']}"

    verified_linked = flask_login.current_user.id in verified_ids
    unverified_linked = flask_login.current_user.id in unverified_ids
    linked = flask_login.current_user.id in linked_ids

    return flask.render_template(
        "other_databases/confirm_identity.html",
        source_database=source_database,
        confirmation_url=confirmation_url,
        cancel_url=cancel_url,
        identity_token=identity_token,
        verified_linked=verified_linked,
        unverified_linked=unverified_linked,
        linked=linked
    )


@frontend.route("/other-databases/validate-identity-token/", methods=["GET"])
def validate_identity_token() -> FlaskResponseT:
    identity_token = flask.request.args.get('token')

    if not identity_token:
        return flask.Response({}, mimetype='application/json', status=400)

    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='identity-token-validation')

    try:
        data = serializer.loads(identity_token, max_age=FEDERATED_IDENTITY_TOKENS_MAX_AGE)
    except itsdangerous.BadSignature:
        return flask.Response({}, mimetype='application/json', status=400)

    return flask.jsonify(data)


@frontend.route("/other-databases/delete-inactive-identities/<int:component_id>", methods=["POST"])
@flask_login.login_required
def delete_inactive_identities(component_id: int) -> FlaskResponseT:
    try:
        component = get_component(component_id)
        fed_ids: list[int] = [int(val) for val in flask.request.form.getlist('local_fed_id')]
    except (errors.ComponentDoesNotExistError, ValueError):
        flask.flash(_('Failed to remove user links.'), 'error')
        return flask.redirect(flask.url_for(".component", component_id=component_id))

    if fed_ids:
        delete_user_links_by_fed_ids(flask_login.current_user.id, component, fed_ids)

    return flask.redirect(flask.url_for(".component", component_id=component_id))


@frontend.route("/other-databases/revoke-active-identities/<int:component_id>", methods=["POST"])
@flask_login.login_required
def revoke_active_identities(component_id: int) -> FlaskResponseT:
    try:
        component = get_component(component_id)
        fed_ids: list[int] = [int(val) for val in flask.request.form.getlist('local_fed_id')]
    except (errors.ComponentDoesNotExistError, ValueError):
        flask.flash(_('Failed to revoke user links.'), 'error')
        return flask.redirect(flask.url_for(".component", component_id=component_id))

    if fed_ids:
        try:
            revoke_user_links_by_fed_ids(flask_login.current_user.id, component.id, fed_ids)
        except errors.NoAuthenticationMethodError:
            flask.flash(_('You must keep at least one federated identity to login or add another authentication method first.'), 'error')

    return flask.redirect(flask.url_for(".component", component_id=component_id))


@frontend.route("/other-databases/edit-eln-identity/<int:eln_import_id>", methods=['POST'])
@flask_login.login_required
def modify_eln_identity(eln_import_id: int) -> FlaskResponseT:
    modify_eln_identity_form = ModifyELNIdentityForm()

    if modify_eln_identity_form.validate_on_submit():
        try:
            fed_identity = get_federated_identity(flask_login.current_user.id, eln_or_fed_user_id=modify_eln_identity_form.eln_user_id.data)
        except errors.FederatedIdentityNotFoundError:
            flask.flash(_('An error occurred: Failed to edit federated identity.'), 'error')
            return flask.redirect(flask.url_for('.eln_import', eln_import_id=eln_import_id))

        if modify_eln_identity_form.type.data == 'remove':
            delete_user_link(flask_login.current_user.id, fed_identity.local_fed_id)
            flask.flash(_('Successfully deleted user link.'), 'success')
        elif modify_eln_identity_form.type.data == 'revoke':
            revoke_user_link(flask_login.current_user.id, fed_identity.local_fed_id)
            flask.flash(_('Successfully revoked user link.'), 'success')
        elif modify_eln_identity_form.type.data == 'enable':
            enable_user_link(flask_login.current_user.id, fed_identity.local_fed_id)
            flask.flash(_('Successfully enabled user link.'), 'success')
    else:
        flask.flash(_('An error occurred: Failed to edit federated identity.'), 'error')
    return flask.redirect(flask.url_for('.eln_import', eln_import_id=eln_import_id))


@frontend.route("/federated-login/sp/<int:component_id>", methods=['POST', 'GET'])
@federated_login_route
def federated_login(component_id: int) -> FlaskResponseT:
    try:
        component = get_component(component_id)
    except errors.ComponentDoesNotExistError:
        return flask.redirect(flask.url_for("frontend.sign_in"))

    if not component.fed_login_available:
        flask.flash(_('The federated login is not available for %(component_name)s.', component_name=component.get_name()), 'error')
        return flask.redirect(flask.url_for("frontend.sign_in"))

    is_shared_device = bool(flask.request.form.get('shared_device'))

    try:
        redirect = sp_login(component, shared_device=is_shared_device)
    except errors.FailedMetadataFetchingError:
        flask.flash(_("%(component_name)s disabled the federated login.", component_name=component.get_name()), 'error')
        return flask.redirect(flask.url_for('frontend.sign_in'))
    except errors.InvalidSAMLRequestError:
        flask.flash(_("Failed to use federated login. Please contact an administrator."), 'error')
        return flask.redirect(flask.url_for('frontend.sign_in'))
    return redirect


@frontend.route("/federated-login/sp/metadata.xml/<string:component>", methods=['GET'])
@federated_login_route
def service_provider_metadata(component: str) -> FlaskResponseT:
    try:
        database = get_component_by_uuid(component)
    except errors.ComponentDoesNotExistError:
        return flask.redirect(flask.url_for("frontend.sign_in"))

    return flask.Response(get_sp_metadata(database), mimetype="text/xml")


@frontend.route("/federated-login/sp/acs", methods=['GET'])
@federated_login_route
def assertion_consumer_service() -> FlaskResponseT:
    args = flask.request.args
    if 'SAMLart' not in args and 'current_saml_artifact' not in flask.session:
        flask.flash(_('An error occurred: Failed to use federated login. Please try again.'), 'error')
        return flask.redirect(flask.url_for('frontend.sign_in'))

    if 'SAMLart' in args:
        saml_artifact = args['SAMLart']
        flask.session['current_saml_artifact'] = args['SAMLart']
    else:
        saml_artifact = flask.session['current_saml_artifact']

    try:
        federated_user_information = resolve_artifact(saml_artifact)
    except errors.MismatchedResponseToRequestError:
        flask.flash(_('An error occurred: The authentication could not be verified. Please try again.'), 'error')
        return flask.redirect(flask.url_for('.sign_in'))

    if 'saml_sso_component' not in flask.session:
        flask.flash(_('Failed to use federated login. Please try again.'), 'error')
        return flask.redirect(flask.url_for("frontend.sign_in"))

    component = get_component_by_uuid(flask.session['saml_sso_component'])

    try:
        fed_user = get_user(int(federated_user_information["uid"]), component.id)
    except errors.UserDoesNotExistError:
        fed_user = create_user(name=None, email=None, type=UserType.FEDERATION_USER, component_id=component.id, fed_id=int(federated_user_information["uid"]))
        set_user_hidden(fed_user.id, hidden=True)

    local_user = get_user_by_federated_user(fed_user.id, login=True)

    if local_user is not None:
        flask_login.login_user(local_user)
        return flask.redirect(flask.url_for('.index'))

    federated_identity_exists = get_active_federated_identity_by_federated_user(federated_user_id=fed_user.id) is not None

    flask.session['saml_sso_user_details'] = {
        "fed_user_id": fed_user.fed_id,
        "component_id": fed_user.component_id,
        "user_email": federated_user_information['email']
    }

    authentication_mails = [auth.login['login'] for auth in Authentication.query.filter_by(type=AuthenticationType.EMAIL).all()]
    user_by_email = User.query.filter_by(email=federated_user_information['email']).first()

    if federated_identity_exists:
        flask.flash(_('The federated user already has a federated identity in this database which is not verified. Please sign in to the corresponding user.'), 'info')
    elif user_by_email or federated_user_information['email'] in authentication_mails:
        flask.flash(_('A user with this email address exists already. If you own this account, please use the sign in.'), 'info')

    create_user_form = FederatedUserCreationForm()
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='federated-identities')
    federated_login_token = serializer.dumps((fed_user.fed_id, fed_user.component_id))

    return flask.render_template(
        "other_databases/federated_login.html",
        component=component,
        has_errors=False,
        federated_user_information=federated_user_information,
        allow_create_user=flask.current_app.config['ENABLE_FEDERATED_LOGIN_CREATE_NEW_USER'] and not federated_identity_exists,
        create_user_form=create_user_form,
        federated_login_token=federated_login_token,
        not_verified=federated_identity_exists
    )


@frontend.route("/federated-login/idp/verify", methods=['GET', 'POST'])
@federated_login_route
@login_required_saml
def federated_login_verify() -> FlaskResponseT:
    form_data: dict[str, typing.Any] = {}
    form_data.update(flask.request.args)
    if 'SAMLRequest' in flask.session:
        form_data.update({
            'SAMLRequest': flask.session['SAMLRequest']
        })
        del flask.session['SAMLRequest']

    try:
        response = process_login(form_data)
    except errors.InvalidSAMLRequestError:
        flask.flash(_('Failed to verify the authentication request. Please try again or contact an administrator. If you are an administrator, follow the <a href="%(docs_url)s">documentation</a>.', docs_url="https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/administrator_guide/federation.html#federated-login"), "error-url")
        return flask.redirect(flask.url_for('.index'))

    if response is None:
        flask.flash(_('Failed to use federated login because of missing metadata. Please contact an administrator. If you are an administrator, follow the <a href="%(docs_url)s">documentation</a>.', docs_url="https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/administrator_guide/federation.html#federated-login"), 'error-url')
        return flask.redirect(flask.url_for('.index'))
    return response


@frontend.route("/federated-login/idp/metadata.xml", methods=['GET'])
@federated_login_route
def identity_provider_metadata() -> FlaskResponseT:
    return flask.Response(get_idp_metadata(), mimetype="text/xml")


@frontend.route("/federated-login/idp/ars", methods=['POST'])
@federated_login_route
def artifact_resolution_service() -> FlaskResponseT:
    try:
        response = eval_artifact_resolve(flask.request.data)
    except Exception:
        return flask.Response("Artifact not found", status=404)
    delete_old_artifacts()
    return flask.Response(response=response['data'], headers=response['headers'])


@frontend.route("/federated-login/sp/finalize", methods=['POST'])
@federated_login_route
def finalize_federated_login() -> FlaskResponseT:
    create_user_form = FederatedUserCreationForm()
    sso_details = flask.session['saml_sso_user_details']

    if create_user_form.validate_on_submit():
        mail_status_send = send_federated_login_account_creation_email(
            email=sso_details['user_email'],
            username=create_user_form.username.data,
            component_id=sso_details['component_id'],
            fed_id=sso_details['fed_user_id']
        )[0]

        if mail_status_send == BackgroundTaskStatus.FAILED:
            flask.flash(_("Sending an email failed. Please try again later or contact an administrator."), 'error')
        else:
            flask.flash(_("Please see your email to confirm this mail address and create the user."), 'success')

    else:
        flask.flash(_('The selected username is not valid: %(error)s', error=create_user_form.username.errors[0]), 'error')
        return flask.redirect(flask.url_for('.assertion_consumer_service'))
    return flask.redirect(flask.url_for('.index'))


@frontend.route("/federated-login/create-account/<string:token>", methods=['GET'])
@federated_login_route
def create_account_by_federated_login(token: str) -> FlaskResponseT:
    serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='federated_login_account')
    try:
        token_data = serializer.loads(token, max_age=flask.current_app.config['INVITATION_TIME_LIMIT'])
    except itsdangerous.BadData:
        flask.flash(_("Invalid token. Please login again."), 'error')
        return flask.redirect(flask.url_for('.index'))

    try:
        federated_user = get_user(user_id=token_data['fed_id'], component_id=token_data['component_id'])
    except errors.UserDoesNotExistError:
        federated_user = None

    if federated_user is not None and get_user_by_federated_user(federated_user_id=federated_user.id):
        flask.flash(_("The federated user already has a local federated identity. Please login with this user."), 'error')
        return flask.redirect(flask.url_for('.index'))

    local_user = create_user(
        name=token_data['username'],
        email=token_data['email'],
        type=UserType.PERSON
    )

    create_sampledb_federated_identity(
        user_id=local_user.id,
        component=get_component(component_id=token_data['component_id']),
        fed_id=token_data['fed_id'],
        use_for_login=True
    )

    flask.flash(_("The user was successfully created. Please sign in."), 'success')
    return flask.render_template('sign_in.html',
                                 saml_user_creation=True,
                                 component=get_component(token_data['component_id']),
                                 )



# coding: utf-8
"""

"""
import datetime
import flask
import flask_login

from .. import frontend
from ...logic.action_types import get_action_types
from ...logic.action_permissions import get_sorted_actions_for_user
from ...logic.users import get_users
from ..utils import get_search_paths
from ...utils import FlaskResponseT


@frontend.route('/objects/search/')
@flask_login.login_required
def search() -> FlaskResponseT:
    actions = get_sorted_actions_for_user(
        user_id=flask_login.current_user.id
    )

    action_types = [
        action_type
        for action_type in get_action_types(filter_fed_defaults=True)
        if action_type.show_in_object_filters
    ]

    search_paths, search_paths_by_action, search_paths_by_action_type = get_search_paths(actions, action_types, include_file_name=True)

    if None in search_paths_by_action_type:
        del search_paths_by_action_type[None]

    return flask.render_template(
        'search.html',
        search_paths=search_paths,
        search_paths_by_action=search_paths_by_action,
        search_paths_by_action_type=search_paths_by_action_type,
        actions=actions,
        action_types=action_types,
        datetime=datetime,
        users=get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN']),
    ), 200, {
        'Cache-Control': 'no-cache, no-store, must-revalidate',
        'Pragma': 'no-cache',
        'Expires': '0'
    }



# coding: utf-8
"""

"""

import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .components import Component
from .users import User
from .utils import Model


class MarkdownImage(Model):
    __tablename__ = 'markdown_images'
    __table_args__ = (
        db.UniqueConstraint('file_name', 'component_id', name='markdown_images_file_name_component_id_key'),
    )

    file_name: Mapped[str] = db.Column(db.Text, nullable=False)
    content: Mapped[bytes] = db.Column(db.LargeBinary, nullable=False)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=True)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    permanent: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey(Component.id), nullable=True)
    id: Mapped[int] = db.Column(db.Integer, primary_key=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["MarkdownImage"]]

    def __init__(
            self,
            file_name: str,
            content: bytes,
            user_id: typing.Optional[int],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            permanent: typing.Optional[bool] = False,
            component_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            file_name=file_name,
            content=content,
            user_id=user_id,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            permanent=permanent,
            component_id=component_id
        )



# coding: utf-8
"""

"""


from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, SelectMultipleField, HiddenField, BooleanField
from wtforms.validators import InputRequired, Length, DataRequired


class EditGroupForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    categories = SelectMultipleField()


class CreateGroupForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    categories = SelectMultipleField()


class LeaveGroupForm(FlaskForm):
    pass


class DeleteGroupForm(FlaskForm):
    pass


class RemoveGroupMemberForm(FlaskForm):
    pass


class InviteUserForm(FlaskForm):
    user_id = IntegerField(validators=[InputRequired()])
    add_directly = BooleanField(default=False)


class ToggleFavoriteActionForm(FlaskForm):
    action_id = IntegerField(validators=[InputRequired()])


class ToggleFavoriteInstrumentForm(FlaskForm):
    instrument_id = IntegerField(validators=[InputRequired()])


class NotificationModeForm(FlaskForm):
    pass


class OtherSettingsForm(FlaskForm):
    pass


class CreateAPITokenForm(FlaskForm):
    description = StringField('description', validators=[Length(min=1, max=100)])


class ManageTwoFactorAuthenticationMethodForm(FlaskForm):
    method_id = IntegerField(validators=[InputRequired()])
    action = StringField(validators=[InputRequired()])


class AddWebhookForm(FlaskForm):
    name = StringField(validators=[Length(min=0, max=100)])
    address = StringField(validators=[Length(min=1, max=100)])


class RemoveWebhookForm(FlaskForm):
    id = IntegerField('Authentication_method_id', validators=[DataRequired()])


class WebAuthnLoginForm(FlaskForm):
    assertion = HiddenField(validators=[DataRequired()])
    remember_me = BooleanField()
    shared_device = BooleanField()



# coding: utf-8
"""
Add missing NOT NULL constraints to columns in download_service_job_files table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('download_service_job_files', 'creation') and not column_is_nullable('download_service_job_files', 'expiration'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE download_service_job_files
            ALTER COLUMN creation SET NOT NULL,
            ALTER COLUMN expiration SET NOT NULL
    """))
    return True



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .objects import Objects
from .utils import Model

if typing.TYPE_CHECKING:
    from ..logic.users import User

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@enum.unique
class ObjectLogEntryType(enum.Enum):
    OTHER = 0
    CREATE_OBJECT = 1
    EDIT_OBJECT = 2
    RESTORE_OBJECT_VERSION = 3
    USE_OBJECT_IN_MEASUREMENT = 4
    POST_COMMENT = 5
    UPLOAD_FILE = 6
    USE_OBJECT_IN_SAMPLE_CREATION = 7
    CREATE_BATCH = 8
    ASSIGN_LOCATION = 9
    LINK_PUBLICATION = 10
    REFERENCE_OBJECT_IN_METADATA = 11
    EXPORT_TO_DATAVERSE = 12
    LINK_PROJECT = 13
    UNLINK_PROJECT = 14
    IMPORT_FROM_ELN_FILE = 15


class ObjectLogEntry(Model):
    __tablename__ = 'object_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[ObjectLogEntryType] = db.Column(db.Enum(ObjectLogEntryType), nullable=False)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    is_imported: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ObjectLogEntry"]]

    def __init__(
            self,
            type: ObjectLogEntryType,
            object_id: int,
            user_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None,
            is_imported: bool = False
    ) -> None:
        super().__init__(
            type=type,
            object_id=object_id,
            user_id=user_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            is_imported=is_imported
        )
        self.user: typing.Optional['User'] = None

    def __eq__(self, other: typing.Any) -> bool:
        return isinstance(other, ObjectLogEntry) and \
            self.id == other.id and \
            self.type == other.type and \
            self.object_id == other.object_id and \
            self.user_id == other.user_id and \
            self.data == other.data and \
            self.utc_datetime == other.utc_datetime

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, object_id={self.object_id}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, data={self.data})>'



import datetime
import typing
from uuid import UUID
from functools import wraps

import flask
import flask_login
from flask_babel import _
import werkzeug

from .. import errors


ResponseContent = typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.Dict[int, typing.Any], typing.List[typing.Any], str, bool, int]]
ResponseData = typing.Union[werkzeug.Response, ResponseContent, typing.Tuple[ResponseContent], typing.Tuple[ResponseContent, int], typing.Tuple[ResponseContent, int, typing.Dict[str, str]]]

FlaskRouteT = typing.TypeVar('FlaskRouteT', bound=typing.Callable[..., typing.Any])


@typing.overload
def _get_id(
        id: typing.Any,
        *,
        min: int = 1,
        special_values: typing.Optional[typing.List[int]] = None,
        convert: bool = True,
        default: int,
        mandatory: bool = True,
) -> int:
    ...


@typing.overload
def _get_id(
        id: typing.Any,
        *,
        min: int = 1,
        special_values: typing.Optional[typing.List[int]] = None,
        convert: bool = True,
        default: typing.Optional[int] = None,
        mandatory: typing.Literal[True] = True
) -> int:
    ...


@typing.overload
def _get_id(
        id: typing.Any,
        *,
        min: int = 1,
        special_values: typing.Optional[typing.List[int]] = None,
        convert: bool = True,
        default: typing.Optional[int] = None,
        mandatory: bool = True
) -> typing.Optional[int]:
    ...


def _get_id(
        id: typing.Any,
        *,
        min: int = 1,
        special_values: typing.Optional[typing.List[int]] = None,
        convert: bool = True,
        default: typing.Optional[int] = None,
        mandatory: bool = True
) -> typing.Optional[int]:
    if id is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing ID')
        return default
    if type(id) is not int:
        if convert:
            try:
                id = int(id)
                # mypy type narrowing
                assert type(id) is int
            except ValueError:
                raise errors.InvalidDataExportError(f'ID "{id}" could not be converted to an integer')
        else:
            raise errors.InvalidDataExportError(f'ID "{id}" is not an integer')
    if min is not None and id < min:
        if special_values is not None and id in special_values:
            return id
        raise errors.InvalidDataExportError(f'Invalid ID "{id}". Has to be greater than {min}. Allowed special values: {special_values}')
    return id


@typing.overload
def _get_uuid(
        uuid: typing.Any,
        *,
        default: str,
        mandatory: bool = True
) -> str:
    ...


@typing.overload
def _get_uuid(
        uuid: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: typing.Literal[True] = True
) -> str:
    ...


@typing.overload
def _get_uuid(
        uuid: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: bool = True
) -> typing.Optional[str]:
    ...


def _get_uuid(
        uuid: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: bool = True
) -> typing.Optional[str]:
    if uuid is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing UUID')
        return default
    if type(uuid) is not str:
        raise errors.InvalidDataExportError(f'UUID "{uuid}" is not a string')
    try:
        uuid_obj = UUID(uuid)
        return str(uuid_obj)
    except ValueError:
        raise errors.InvalidDataExportError(f'Invalid UUID "{uuid}"')
    except TypeError:
        raise errors.InvalidDataExportError(f'Invalid UUID "{uuid}"')


@typing.overload
def _get_bool(
        bool_in: typing.Any,
        *,
        default: bool,
        mandatory: bool = False
) -> bool:
    ...


@typing.overload
def _get_bool(
        bool_in: typing.Any,
        *,
        default: typing.Optional[bool] = None,
        mandatory: typing.Literal[True]
) -> bool:
    ...


@typing.overload
def _get_bool(
        bool_in: typing.Any,
        *,
        default: typing.Optional[bool] = None,
        mandatory: bool = False
) -> typing.Optional[bool]:
    ...


def _get_bool(
        bool_in: typing.Any,
        *,
        default: typing.Optional[bool] = None,
        mandatory: bool = False
) -> typing.Optional[bool]:
    if bool_in is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing boolean')
        return default
    if type(bool_in) is not bool:
        raise errors.InvalidDataExportError(f'Invalid boolean "{bool_in}"')
    return bool_in


@typing.overload
def _get_int(
        int_in: typing.Any,
        *,
        default: int,
        mandatory: bool = False
) -> int:
    ...


@typing.overload
def _get_int(
        int_in: typing.Any,
        *,
        default: typing.Optional[int] = None,
        mandatory: typing.Literal[True]
) -> int:
    ...


@typing.overload
def _get_int(
        int_in: typing.Any,
        *,
        default: typing.Optional[int] = None,
        mandatory: bool = False
) -> typing.Optional[int]:
    ...


def _get_int(
        int_in: typing.Any,
        *,
        default: typing.Optional[int] = None,
        mandatory: bool = False
) -> typing.Optional[int]:
    if int_in is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing integer')
        return default
    if type(int_in) is not int:
        raise errors.InvalidDataExportError(f'Invalid integer "{int_in}"')
    return int_in


def _get_translation(
        translation: typing.Any,
        *,
        default: typing.Optional[typing.Dict[str, str]] = None,
        mandatory: bool = False
) -> typing.Optional[typing.Dict[str, str]]:
    if translation is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing text')
        return default
    if isinstance(translation, dict):
        for key, item in translation.items():
            if type(key) is not str or type(item) is not str or key == '':
                raise errors.InvalidDataExportError(f'Invalid translation dict "{translation}"')
        return translation
    if type(translation) is not str:
        raise errors.InvalidDataExportError(f'Text is neither a dictionary nor string "{translation}"')
    return {'en': translation}


@typing.overload
def _get_dict(
        dict_in: typing.Any,
        *,
        default: typing.Dict[typing.Any, typing.Any],
        mandatory: bool = False
) -> typing.Dict[typing.Any, typing.Any]:
    ...


@typing.overload
def _get_dict(
        dict_in: typing.Any,
        *,
        default: None = None,
        mandatory: typing.Literal[True] = True
) -> typing.Dict[typing.Any, typing.Any]:
    ...


@typing.overload
def _get_dict(
        dict_in: typing.Any,
        *,
        default: None = None,
        mandatory: bool = False
) -> typing.Optional[typing.Dict[typing.Any, typing.Any]]:
    ...


def _get_dict(
        dict_in: typing.Any,
        *,
        default: typing.Optional[typing.Dict[typing.Any, typing.Any]] = None,
        mandatory: bool = False
) -> typing.Optional[typing.Dict[typing.Any, typing.Any]]:
    if dict_in is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing dict')
        return default
    if not isinstance(dict_in, dict):
        raise errors.InvalidDataExportError(f'Invalid dict "{dict_in}"')
    return dict_in


def _get_permissions(
        permissions_in: typing.Any,
        *,
        default: typing.Optional[typing.Dict[str, typing.Union[str, typing.Dict[int, str]]]] = None
) -> typing.Optional[typing.Dict[str, typing.Union[str, typing.Dict[int, str]]]]:
    if permissions_in is None:
        return default
    permissions = _get_dict(permissions_in, mandatory=True)
    users = _get_dict(permissions.get('users'), default={})
    groups = _get_dict(permissions.get('groups'), default={})
    projects = _get_dict(permissions.get('projects'), default={})
    _get_str(permissions.get('all_users'), default='none')
    for id, perm in users.items():
        _get_id(id, mandatory=True)
        _get_str(perm, mandatory=True)
    for id, perm in groups.items():
        _get_id(id, mandatory=True)
        _get_str(perm, mandatory=True)
    for id, perm in projects.items():
        _get_id(id, mandatory=True)
        _get_str(perm, mandatory=True)
    return permissions


@typing.overload
def _get_list(
        list_in: typing.Any,
        *,
        default: typing.List[typing.Any],
        mandatory: bool = False
) -> typing.List[typing.Any]:
    ...


@typing.overload
def _get_list(
        list_in: typing.Any,
        *,
        default: typing.Optional[typing.List[typing.Any]] = None,
        mandatory: typing.Literal[True]
) -> typing.List[typing.Any]:
    ...


@typing.overload
def _get_list(
        list_in: typing.Any,
        *,
        default: typing.Optional[typing.List[typing.Any]] = None,
        mandatory: bool = False
) -> typing.Optional[typing.List[typing.Any]]:
    ...


def _get_list(
        list_in: typing.Any,
        *,
        default: typing.Optional[typing.List[typing.Any]] = None,
        mandatory: bool = False
) -> typing.Optional[typing.List[typing.Any]]:
    if list_in is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing list')
        return default
    if not isinstance(list_in, list):
        raise errors.InvalidDataExportError(f'Invalid list "{list_in}"')
    return list_in


@typing.overload
def _get_str(
        str_in: typing.Any,
        *,
        default: str,
        mandatory: bool = False,
        allow_empty: bool = True,
        convert: bool = False
) -> str:
    ...


@typing.overload
def _get_str(
        str_in: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: typing.Literal[True],
        allow_empty: bool = True,
        convert: bool = False
) -> str:
    ...


@typing.overload
def _get_str(
        str_in: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: bool = False,
        allow_empty: bool = True,
        convert: bool = False
) -> typing.Optional[str]:
    ...


def _get_str(
        str_in: typing.Any,
        *,
        default: typing.Optional[str] = None,
        mandatory: bool = False,
        allow_empty: bool = True,
        convert: bool = False
) -> typing.Optional[str]:
    if str_in is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing string')
        return default
    if type(str_in) is not str:
        if not convert:
            raise errors.InvalidDataExportError(f'"{str_in}" is not a string')
        try:
            str_in = str(str_in)
            # mypy type narrowing
            assert type(str_in) is str
        except ValueError:
            raise errors.InvalidDataExportError(f'Cannot convert "{str_in}" to string')
    if not allow_empty and str_in == '':
        raise errors.InvalidDataExportError('Empty string')
    return str_in


@typing.overload
def _get_utc_datetime(
        utc_datetime_str: typing.Any,
        *,
        default: datetime.datetime,
        mandatory: bool = False
) -> datetime.datetime:
    ...


@typing.overload
def _get_utc_datetime(
        utc_datetime_str: typing.Any,
        *,
        default: typing.Optional[datetime.datetime] = None,
        mandatory: typing.Literal[True]
) -> datetime.datetime:
    ...


@typing.overload
def _get_utc_datetime(
        utc_datetime_str: typing.Any,
        *,
        default: typing.Optional[datetime.datetime] = None,
        mandatory: bool = False
) -> typing.Optional[datetime.datetime]:
    ...


def _get_utc_datetime(
        utc_datetime_str: typing.Any,
        *,
        default: typing.Optional[datetime.datetime] = None,
        mandatory: bool = False
) -> typing.Optional[datetime.datetime]:
    if utc_datetime_str is None:
        if mandatory:
            raise errors.InvalidDataExportError('Missing timestamp')
        return default
    try:
        dt = datetime.datetime.strptime(utc_datetime_str, '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=datetime.timezone.utc)
        if dt > datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=flask.current_app.config['VALID_TIME_DELTA']):
            raise errors.InvalidDataExportError(f'Timestamp is in the future "{dt}"')
        return dt
    except ValueError:
        raise errors.InvalidDataExportError(f'Invalid timestamp "{utc_datetime_str}"')
    except TypeError:
        raise errors.InvalidDataExportError(f'Invalid timestamp "{utc_datetime_str}"')


def login_required_saml(f: FlaskRouteT) -> FlaskRouteT:
    @wraps(f)
    def decorated(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if flask_login.current_user.is_authenticated:
            return f(*args, **kwargs)
        flask.session['SAMLRequest'] = flask.request.args.get("SAMLRequest", None)
        return flask.current_app.login_manager.unauthorized()  # type: ignore[attr-defined]

    return typing.cast(FlaskRouteT, decorated)


def federated_login_route(f: FlaskRouteT) -> FlaskRouteT:
    @wraps(f)
    def decorated(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if flask.current_app.config['ENABLE_FEDERATED_LOGIN']:
            return f(*args, **kwargs)
        flask.flash(_("Federated login is disabled for %(service_name)s.", service_name=flask.current_app.config['SERVICE_NAME']), 'error')
        return flask.redirect(flask.url_for('frontend.index'))
    return typing.cast(FlaskRouteT, decorated)



# coding: utf-8
"""

"""

import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .actions import SciCatExportType
from .objects import Objects
from .users import User
from .utils import Model


class SciCatExport(Model):
    __tablename__ = 'scicat_exports'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), primary_key=True)
    scicat_url: Mapped[str] = db.Column(db.String, nullable=False)
    scicat_pid: Mapped[str] = db.Column(db.String, nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    type: Mapped[SciCatExportType] = db.Column(db.Enum(SciCatExportType), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["SciCatExport"]]

    def __init__(
            self,
            object_id: int,
            scicat_url: str,
            scicat_pid: str,
            user_id: int,
            type: SciCatExportType,
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            object_id=object_id,
            scicat_url=scicat_url,
            scicat_pid=scicat_pid,
            user_id=user_id,
            type=type,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(object_id={self.object_id}, scicat_url={self.scicat_url})>'



import collections
import dataclasses
import datetime
import typing

import flask

from .. import db, models
from . import errors, users


@dataclasses.dataclass(frozen=True)
class InfoPage:
    """
    This class provides an immutable wrapper around models.info_pages.InfoPage.
    """
    id: int
    title: typing.Dict[str, str]
    content: typing.Dict[str, str]
    endpoint: typing.Optional[str]
    disabled: bool

    @classmethod
    def from_database(cls, info_page: models.InfoPage) -> 'InfoPage':
        return cls(
            id=info_page.id,
            title=info_page.title,
            content=info_page.content,
            endpoint=info_page.endpoint,
            disabled=info_page.disabled,
        )


def get_info_pages() -> typing.Sequence[InfoPage]:
    """
    Get all info pages.

    :return: the list of all info pages
    """
    return [
        InfoPage.from_database(info_page)
        for info_page in models.InfoPage.query.order_by(models.InfoPage.id).all()
    ]


def get_info_page(info_page_id: int) -> InfoPage:
    """
    Get an info page with a given info page ID.

    :param info_page_id: the info page ID
    :return: the info page
    :raise errors.InfoPageDoesNotExistError: if no info page with the given ID
        exists
    """
    info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
    if info_page is None:
        raise errors.InfoPageDoesNotExistError()
    return InfoPage.from_database(info_page)


def get_info_pages_for_endpoint(
        endpoint: typing.Optional[str],
        user_id: typing.Optional[int],
        exclude_disabled: bool = True
) -> typing.Sequence[InfoPage]:
    """
    Get all unacknowledged info pages for an endpoint.

    :param endpoint: the endpoint to get info pages for
    :param user_id: the user id to get info pages for
    :param exclude_disabled: whether to exclude disabled info pages
    :return: a list of info pages
    """
    info_page_query = models.info_pages.InfoPage.query.filter(
        db.or_(
            models.InfoPage.endpoint == endpoint,
            models.InfoPage.endpoint == db.null(),
        )
    )
    if exclude_disabled:
        info_page_query = info_page_query.filter(models.InfoPage.disabled == db.false())
    if user_id is not None:
        info_page_query = info_page_query.where(db.not_(models.InfoPageAcknowledgement.query.filter(
            db.and_(
                models.InfoPageAcknowledgement.info_page_id == models.info_pages.InfoPage.id,
                models.InfoPageAcknowledgement.user_id == user_id,
            )
        ).exists()))
    return [
        InfoPage.from_database(info_page)
        for info_page in info_page_query.all()
    ]


def acknowledge_info_pages(
        info_page_ids: typing.Set[int],
        user_id: int,
) -> None:
    """
    Create acknowledgements for a user for a set of info pages.

    :param info_page_ids: a set of info page IDs to create acknowledgements for
    :param user_id: the user to create the acknowledgements for
    :raise errors.InfoPageDoesNotExistError: if no info page with one of the
        given IDs exists
    :raise errors.UserDoesNotExistError: if no user with the given ID exists
    """
    users.check_user_exists(user_id)
    existing_acknowledgements = {
        acknowledgement.info_page_id: acknowledgement
        for acknowledgement in models.InfoPageAcknowledgement.query.filter(
            db.and_(
                models.InfoPageAcknowledgement.user_id == user_id,
                models.InfoPageAcknowledgement.info_page_id.in_(info_page_ids),
            )
        ).all()
    }
    utc_datetime = datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None)
    all_info_pages = get_info_pages()
    all_info_page_ids = {info_page.id for info_page in all_info_pages}
    for info_page_id in info_page_ids:
        if info_page_id not in all_info_page_ids:
            raise errors.InfoPageDoesNotExistError()
        if info_page_id not in existing_acknowledgements:
            db.session.add(models.InfoPageAcknowledgement(
                info_page_id=info_page_id,
                user_id=user_id,
                utc_datetime=utc_datetime
            ))
        elif existing_acknowledgements[info_page_id].utc_datetime is None:
            existing_acknowledgements[info_page_id].utc_datetime = utc_datetime
            db.session.add(existing_acknowledgements[info_page_id])
    db.session.commit()


def acknowledge_info_page_for_existing_users(
        info_page_id: int
) -> None:
    """
    Create pseudo-acknowledgements for all existing users for an info page.

    :param info_page_id: the ID of the info page to create
        pseudo-acknowledgements for
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
    if info_page is None:
        raise errors.InfoPageDoesNotExistError()
    existing_acknowledgements = {
        acknowledgement.user_id
        for acknowledgement in models.InfoPageAcknowledgement.query.filter_by(info_page_id=info_page_id).all()
    }
    users = models.User.query.all()
    for user in users:
        if user.id in existing_acknowledgements:
            continue
        db.session.add(models.InfoPageAcknowledgement(
            info_page_id=info_page_id,
            user_id=user.id,
            utc_datetime=None
        ))
    db.session.commit()


def clear_acknowledgements_for_info_page(
        info_page_id: int
) -> None:
    """
    Clear all (real) user acknowledgements for an info page.

    This will not clear pseudo-acknowledgements that were created for
    existing users when the info page was created.

    :param info_page_id: the ID of the info page to clean acknowledgements for
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    acknowledgements = models.InfoPageAcknowledgement.query.filter(
        db.and_(
            models.InfoPageAcknowledgement.info_page_id == info_page_id,
            models.InfoPageAcknowledgement.utc_datetime != db.null(),
        )
    ).all()
    if acknowledgements:
        for acknowledgement in acknowledgements:
            db.session.delete(acknowledgement)
        db.session.commit()
    else:
        info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
        if info_page is None:
            raise errors.InfoPageDoesNotExistError()


def get_acknowledgements_for_info_page(
        info_page_id: int
) -> typing.Dict[int, typing.Optional[datetime.datetime]]:
    """
    Get user acknowledgements for an info page ID.

    :param info_page_id: the ID of the info page to get acknowledgements for
    :return: a dict mapping user IDs to their acknowledgements for the info page
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    acknowledgements = models.InfoPageAcknowledgement.query.filter_by(info_page_id=info_page_id).all()
    if not acknowledgements:
        info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
        if info_page is None:
            raise errors.InfoPageDoesNotExistError()
        return {}
    return {
        acknowledgement.user_id: acknowledgement.utc_datetime
        for acknowledgement in acknowledgements
    }


def create_info_page(
        title: typing.Dict[str, str],
        content: typing.Dict[str, str],
        endpoint: typing.Optional[str],
        disabled: bool = False,
) -> InfoPage:
    """
    Create a new info page.

    :param title: the title for the info page
    :param content: the content for the info page
    :param endpoint: the endpoint for the info page
    :param disabled: whether the info page should be disabled
    :return: the created info page
    """
    info_page = models.InfoPage(
        title=title,
        content=content,
        endpoint=endpoint,
        disabled=disabled,
    )
    db.session.add(info_page)
    db.session.commit()
    return InfoPage.from_database(info_page)


def set_info_page_disabled(
        info_page_id: int,
        disabled: bool = True
) -> None:
    """
    Disable or enable an info page.

    :param info_page_id: the ID of the info page to disable or enable
    :param disabled: whether the info page should be disabled
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
    if info_page is None:
        raise errors.InfoPageDoesNotExistError()
    info_page.disabled = disabled
    db.session.add(info_page)
    db.session.commit()


def update_info_page(
        info_page_id: int,
        title: typing.Dict[str, str],
        content: typing.Dict[str, str],
        endpoint: typing.Optional[str],
) -> InfoPage:
    """
    Update an info page.

    :param info_page_id: the ID of the info page to update
    :param title: the new title for the info page
    :param content: the new content for the info page
    :param endpoint: the new endpoint for the info page
    :return: the updated info page
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
    if info_page is None:
        raise errors.InfoPageDoesNotExistError()
    info_page.title = title
    info_page.content = content
    info_page.endpoint = endpoint
    db.session.add(info_page)
    db.session.commit()
    return InfoPage.from_database(info_page)


def delete_info_page(
        info_page_id: int
) -> None:
    """
    Delete an info page.

    :param info_page_id: the ID of the info page to delete
    :raise errors.InfoPageDoesNotExistError: if no info page with the given
        info page ID exists
    """
    info_page = models.InfoPage.query.filter_by(id=info_page_id).first()
    if info_page is None:
        raise errors.InfoPageDoesNotExistError()
    db.session.delete(info_page)
    db.session.commit()


def get_url_rules_by_endpoint() -> typing.Dict[str, typing.List[str]]:
    """
    Get the list of URL rules that are reasonable for info pages.

    :return: the list of rules

    While info pages can be created for any endpoint, info pages will not
    work with many endpoints, e.g. because they cannot be accessed with GET or
    because they will not use the base template that displays info pages. This
    function tries to filter out these endpoints that would not be useful for
    displaying an info page.
    """
    url_rules_by_endpoint = collections.defaultdict(list)
    for url_rule in flask.current_app.url_map.iter_rules():
        # info pages can only be shown via GET
        if url_rule.methods and 'GET' not in url_rule.methods:
            continue
        # info pages only work in the frontend
        if not url_rule.endpoint.startswith('frontend.'):
            continue
        # info pages do not work during sign in
        if url_rule.rule.startswith('/users/me/sign_in'):
            continue
        if url_rule.rule.startswith('/users/me/two_factor_authentication/'):
            continue
        if url_rule.rule.startswith('/federated-login/'):
            continue
        if url_rule.rule.startswith('/users/me/oidc'):
            continue
        if url_rule.rule.startswith('/other-databases/') and 'identity' in url_rule.rule:
            continue
        # info pages do not work for files
        if '.' in url_rule.rule:
            continue
        if '/file_attachments/' in url_rule.rule:
            continue
        if url_rule.rule.startswith('/markdown_images/'):
            continue
        if url_rule.rule.startswith('/objects/') and (url_rule.rule.endswith('/export') or url_rule.rule.endswith('/label')):
            continue
        if url_rule.rule == '/multiselect_labels':
            continue
        # info pages do not work on routes that are simple redirects
        if url_rule.rule.endswith('/activity'):
            continue
        if url_rule.rule.startswith('/locations/') and url_rule.rule.endswith('_responsibility'):
            continue
        if '/redirect-uuid/' in url_rule.rule:
            continue
        if url_rule.rule == '/users/me':
            continue
        # info pages do not work on routes that do not use base.html
        if '/mobile_upload/' in url_rule.rule:
            continue
        if url_rule.rule in {'/objects/referencable', '/status/', '/users/me/shared_device_state'}:
            continue
        url_rules_by_endpoint[url_rule.endpoint].append(url_rule.rule)
    return url_rules_by_endpoint



import copy
import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_list, _get_dict, _get_translation
from .location_types import _parse_location_type_ref, _get_or_create_location_type_id, LocationTypeRef
from .users import _parse_user_ref, _get_or_create_user_id, import_user, UserRef, UserData
from ..locations import LocationType, get_location_type, set_location_responsible_users, get_location, get_locations, create_location, update_location, Location
from ..components import Component, get_component, get_component_by_uuid
from .. import errors, fed_logs, location_permissions
from ...models import Permissions


class LocationRef(typing.TypedDict):
    location_id: int
    component_uuid: str


class LocationPermissionsData(typing.TypedDict):
    all_users: typing.Union[typing.Literal['none'], typing.Literal['read']]


class LocationData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    name: typing.Optional[typing.Dict[str, str]]
    description: typing.Optional[typing.Dict[str, str]]
    parent_location: typing.Optional[LocationRef]
    location_type: typing.Optional[LocationTypeRef]
    responsible_users: typing.List[UserRef]
    permissions: typing.Optional[LocationPermissionsData]


class SharedLocationData(typing.TypedDict):
    location_id: int
    component_uuid: str
    name: typing.Optional[typing.Dict[str, str]]
    description: typing.Optional[typing.Dict[str, str]]
    parent_location: typing.Optional[LocationRef]
    location_type: typing.Optional[LocationTypeRef]
    responsible_users: typing.List[UserRef]
    permissions: typing.Optional[LocationPermissionsData]


def parse_location(
        location_data: typing.Dict[str, typing.Any]
) -> LocationData:
    fed_id = _get_id(location_data.get('location_id'))
    uuid = _get_uuid(location_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local location {fed_id}')
    responsible_users = []
    for responsible_user_data in _get_list(location_data.get('responsible_users'), default=[]):
        responsible_user_ref = _parse_user_ref(_get_dict(responsible_user_data))
        if responsible_user_ref is not None:
            responsible_users.append(responsible_user_ref)
    return LocationData(
        fed_id=fed_id,
        component_uuid=uuid,
        name=_get_translation(location_data.get('name')),
        description=_get_translation(location_data.get('description')),
        parent_location=_parse_location_ref(_get_dict(location_data.get('parent_location'))),
        location_type=_parse_location_type_ref(_get_dict(location_data.get('location_type'))),
        responsible_users=responsible_users,
        permissions=_parse_location_permissions(_get_dict(location_data.get('permissions'))),
    )


def import_location(
        location_data: LocationData,
        component: Component,
        locations: typing.Dict[typing.Tuple[int, str], LocationData],
        users: typing.List[UserData]
) -> Location:
    component_id = _get_or_create_component_id(location_data['component_uuid'])
    # component_id will only be None if this would import a local location
    assert component_id is not None

    parent = _parse_location_ref(location_data.get('parent_location'))
    if parent is None:
        parent_location_id = None
    else:
        if (parent['location_id'], parent['component_uuid']) in locations.keys():
            parent_location_id = import_location(locations[(parent['location_id'], parent['component_uuid'])], component, locations, users).id
            locations.pop((location_data['fed_id'], location_data['component_uuid']))
        else:
            parent_location_id = _get_or_create_location_id(location_data['parent_location'])
    location_type = _parse_location_type_ref(location_data.get('location_type'))
    if location_type is None:
        location_type_id = LocationType.LOCATION
    else:
        location_type_id = _get_or_create_location_type_id(location_type)
    responsible_user_ids = []
    if location_data.get('responsible_users'):
        users_by_ids = {
            (user['fed_id'], user['component_uuid']): user
            for user in users
        }
        for responsible_user_data in location_data['responsible_users']:
            responsible_user = _parse_user_ref(responsible_user_data)
            if responsible_user is None:
                responsible_user_id = None
            else:
                if (responsible_user['user_id'], responsible_user['component_uuid']) in users_by_ids.keys():
                    responsible_user_id = import_user(users_by_ids[(responsible_user['user_id'], responsible_user['component_uuid'])], component).id
                else:
                    responsible_user_id = _get_or_create_user_id(responsible_user_data)
            if responsible_user_id is not None:
                responsible_user_ids.append(responsible_user_id)
    try:
        location = get_location(location_data['fed_id'], component_id)
        if location.name != location_data['name'] or location.description != location_data['description'] or location.parent_location_id != parent_location_id:
            update_location(
                location_id=location.id,
                name=location_data['name'],
                description=location_data['description'],
                parent_location_id=parent_location_id,
                user_id=None,
                type_id=location_type_id,
                is_hidden=location.is_hidden,
                enable_object_assignments=location.enable_object_assignments
            )
            set_location_responsible_users(
                location_id=location.id,
                responsible_user_ids=responsible_user_ids
            )
            fed_logs.update_location(location.id, component.id)
    except errors.LocationDoesNotExistError:
        location = create_location(
            fed_id=location_data['fed_id'],
            component_id=component_id,
            name=location_data['name'],
            description=location_data['description'],
            parent_location_id=parent_location_id,
            user_id=None,
            type_id=location_type_id
        )
        set_location_responsible_users(
            location_id=location.id,
            responsible_user_ids=responsible_user_ids
        )
        fed_logs.import_location(location.id, component.id)
    location_permissions_data = location_data.get('permissions')
    if location_permissions_data is not None and location_permissions_data.get('all_users') == 'read':
        location_permissions.set_location_permissions_for_all_users(location.id, Permissions.READ)
    return location


def parse_import_location(
        location_data: typing.Dict[str, typing.Any],
        component: Component
) -> Location:
    parsed_location_data = parse_location(location_data)
    locations = {
        (parsed_location_data['fed_id'], parsed_location_data['component_uuid']): parsed_location_data
    }
    locations_check_for_cyclic_dependencies(locations)
    return import_location(parsed_location_data, component, locations, [])


def _parse_location_ref(
        location_data: typing.Optional[typing.Union[LocationRef, typing.Dict[str, typing.Any]]]
) -> typing.Optional[LocationRef]:
    if location_data is None:
        return None
    location_id = _get_id(location_data.get('location_id'))
    component_uuid = _get_uuid(location_data.get('component_uuid'))
    return LocationRef(
        location_id=location_id,
        component_uuid=component_uuid
    )


def _parse_location_permissions(
        location_permissions_data: typing.Optional[typing.Dict[str, typing.Any]]
) -> typing.Optional[LocationPermissionsData]:
    if location_permissions_data is None:
        return None
    all_users_permissions = location_permissions_data.get('all_users')
    if all_users_permissions not in ('none', 'read'):
        return None
    return LocationPermissionsData(
        all_users=all_users_permissions,
    )


def _get_or_create_location_id(
        location_data: typing.Optional[LocationRef]
) -> typing.Optional[int]:
    if location_data is None:
        return None
    component_id = _get_or_create_component_id(location_data['component_uuid'])
    try:
        location = get_location(location_data['location_id'], component_id)
    except errors.LocationDoesNotExistError:
        assert component_id is not None
        location = create_location(
            name=None,
            description=None,
            parent_location_id=None,
            user_id=None,
            type_id=LocationType.LOCATION,
            fed_id=location_data['location_id'],
            component_id=component_id
        )
        fed_logs.create_ref_location(location.id, component_id)
    return location.id


def shared_location_preprocessor(
        location_id: int,
        _component: Component,
        refs: typing.List[typing.Tuple[str, int]],
        _markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedLocationData]:
    location = get_location(location_id)
    if location.component_id is not None:
        return None
    if location.parent_location_id is not None:
        if ('locations', location.parent_location_id) not in refs:
            refs.append(('locations', location.parent_location_id))
        parent = get_location(location.parent_location_id)
        if parent.component is None or parent.fed_id is None:
            parent_location = LocationRef(
                location_id=parent.id,
                component_uuid=flask.current_app.config['FEDERATION_UUID']
            )
        else:
            comp = get_component(parent.component.id)
            parent_location = LocationRef(
                location_id=parent.fed_id,
                component_uuid=comp.uuid
            )
    else:
        parent_location = None
    if location.type_id is not None:
        if ('location_types', location.type_id) not in refs:
            refs.append(('location_types', location.type_id))
        location_type = get_location_type(location.type_id)
        if location_type.component is None or location_type.fed_id is None:
            location_type_ref = LocationTypeRef(
                location_type_id=location_type.id,
                component_uuid=flask.current_app.config['FEDERATION_UUID']
            )
        else:
            location_type_ref = LocationTypeRef(
                location_type_id=location_type.fed_id,
                component_uuid=location_type.component.uuid
            )
    else:
        location_type_ref = None
    responsible_users: typing.List[UserRef] = []
    if location.responsible_users:
        for responsible_user in location.responsible_users:
            if ('users', responsible_user.id) not in refs:
                refs.append(('users', responsible_user.id))
            if responsible_user.component is None or responsible_user.fed_id is None:
                responsible_user_ref = UserRef(
                    user_id=responsible_user.id,
                    component_uuid=flask.current_app.config['FEDERATION_UUID']
                )
            else:
                responsible_user_ref = UserRef(
                    user_id=responsible_user.fed_id,
                    component_uuid=responsible_user.component.uuid
                )
            responsible_users.append(responsible_user_ref)

    location_permissions_data = LocationPermissionsData(
        all_users='read' if location_permissions.location_is_public(location.id) else 'none'
    )
    return SharedLocationData(
        location_id=location.id if location.fed_id is None else location.fed_id,
        component_uuid=flask.current_app.config['FEDERATION_UUID'] if location.component is None else location.component.uuid,
        name=location.name,
        description=location.description,
        parent_location=parent_location,
        location_type=location_type_ref,
        responsible_users=responsible_users,
        permissions=location_permissions_data
    )


def locations_check_for_cyclic_dependencies(
        locations_data: typing.Dict[typing.Tuple[int, str], LocationData]
) -> None:
    def _location_check_for_cyclic_dependencies(
            location_data: LocationData,
            locations: typing.Dict[typing.Tuple[int, str], LocationData],
            path: typing.List[typing.Tuple[int, str]]
    ) -> None:
        if (location_data['fed_id'], location_data['component_uuid']) in path:
            # cyclic parent-location-chain
            raise errors.InvalidDataExportError(f'Cyclic parent location chain. Location: #{location_data["fed_id"]} @ {location_data["component_uuid"]} Path: {path}')
        if location_data['parent_location'] is not None:
            if (location_data['parent_location']['location_id'], location_data['parent_location']['component_uuid']) in locations.keys():
                _location_check_for_cyclic_dependencies(
                    locations[(location_data['parent_location']['location_id'], location_data['parent_location']['component_uuid'])],
                    locations, path + [(location_data['fed_id'], location_data['component_uuid'])]
                )
            else:
                try:
                    component = get_component_by_uuid(location_data['component_uuid'])
                except errors.ComponentDoesNotExistError:
                    return
                try:
                    get_location(location_data['fed_id'], component.id)
                except errors.LocationDoesNotExistError:
                    return

    data = copy.deepcopy(locations_data)
    local_locations = get_locations()
    for location in local_locations:
        if location.component is not None and location.fed_id is not None:
            if (location.fed_id, location.component.uuid) in data.keys():
                continue
            if location.parent_location_id is None:
                data[(location.fed_id, location.component.uuid)] = LocationData(
                    fed_id=location.fed_id,
                    component_uuid=location.component.uuid,
                    parent_location=None,
                    name=None,
                    description=None,
                    location_type=None,
                    responsible_users=[],
                    permissions={'all_users': 'none'},
                )
            else:
                parent_location = get_location(location.parent_location_id)
                if parent_location.component is not None and parent_location.fed_id is not None:
                    data[(location.fed_id, location.component.uuid)] = LocationData(
                        fed_id=location.fed_id,
                        component_uuid=location.component.uuid,
                        parent_location=LocationRef(
                            location_id=parent_location.fed_id,
                            component_uuid=parent_location.component.uuid
                        ),
                        name=None,
                        description=None,
                        location_type=None,
                        responsible_users=[],
                        permissions={'all_users': 'none'},
                    )
                else:
                    data[(location.fed_id, location.component.uuid)] = LocationData(
                        fed_id=location.fed_id,
                        component_uuid=location.component.uuid,
                        parent_location=LocationRef(
                            location_id=parent_location.id,
                            component_uuid=flask.current_app.config['FEDERATION_UUID']
                        ),
                        name=None,
                        description=None,
                        location_type=None,
                        responsible_users=[],
                        permissions={'all_users': 'none'},
                    )

        else:
            if location.parent_location_id is None:
                data[(location.id, flask.current_app.config['FEDERATION_UUID'])] = LocationData(
                    fed_id=location.id,
                    component_uuid=flask.current_app.config['FEDERATION_UUID'],
                    parent_location=None,
                    name=None,
                    description=None,
                    location_type=None,
                    responsible_users=[],
                    permissions={'all_users': 'none'},
                )
            else:
                parent_location = get_location(location.parent_location_id)
                if parent_location.component is not None and parent_location.fed_id is not None:
                    data[(location.id, flask.current_app.config['FEDERATION_UUID'])] = LocationData(
                        fed_id=location.id,
                        component_uuid=flask.current_app.config['FEDERATION_UUID'],
                        parent_location=LocationRef(
                            location_id=parent_location.fed_id,
                            component_uuid=parent_location.component.uuid
                        ),
                        name=None,
                        description=None,
                        location_type=None,
                        responsible_users=[],
                        permissions={'all_users': 'none'},
                    )
                else:
                    data[(location.id, flask.current_app.config['FEDERATION_UUID'])] = LocationData(
                        fed_id=location.id,
                        component_uuid=flask.current_app.config['FEDERATION_UUID'],
                        parent_location=LocationRef(
                            location_id=parent_location.id,
                            component_uuid=flask.current_app.config['FEDERATION_UUID']
                        ),
                        name=None,
                        description=None,
                        location_type=None,
                        responsible_users=[],
                        permissions={'all_users': 'none'},
                    )

    for _, location_data in data.items():
        _location_check_for_cyclic_dependencies(location_data, data, [])



# coding: utf-8
"""
Logic module for comments

Users with WRITE permissions can comment on samples or measurements. These
comments are immutable and therefore this module only allows the creation and
querying of comments.
"""

import dataclasses
import datetime
import typing

from .. import db, models
from . import user_log, object_log, objects, users, errors, components


@dataclasses.dataclass(frozen=True)
class Comment:
    """
    This class provides an immutable wrapper around models.comments.Comment.
    """
    id: int
    object_id: int
    user_id: typing.Optional[int]
    author: typing.Optional[users.User]
    content: str
    utc_datetime: typing.Optional[datetime.datetime]
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    component: typing.Optional[components.Component] = None

    @classmethod
    def from_database(cls, comment: models.Comment) -> 'Comment':
        return Comment(
            id=comment.id,
            object_id=comment.object_id,
            user_id=comment.user_id,
            author=users.User.from_database(comment.author) if comment.author is not None else None,
            content=comment.content,
            utc_datetime=comment.utc_datetime,
            fed_id=comment.fed_id,
            component_id=comment.component_id,
            component=components.Component.from_database(comment.component) if comment.component is not None else None
        )


@typing.overload
def create_comment(
        object_id: int,
        user_id: int,
        content: str,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        *,
        create_log_entry: bool = True,
        fed_id: None = None,
        component_id: None = None
) -> int:
    ...


@typing.overload
def create_comment(
        object_id: int,
        user_id: typing.Optional[int],
        content: str,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        *,
        create_log_entry: bool = True,
        fed_id: int,
        component_id: int
) -> int:
    ...


def create_comment(
        object_id: int,
        user_id: typing.Optional[int],
        content: str,
        utc_datetime: typing.Optional[datetime.datetime] = None,
        *,
        create_log_entry: bool = True,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None
) -> int:
    """
    Creates a new comment and adds it to the object and user logs.

    :param object_id: the ID of an existing object
    :param user_id: the ID of an existing user
    :param content: the text content for the new comment
    :param utc_datetime: the creation time of the comment or None to select the current time
    :param create_log_entry: whether to create a log entry
    :param fed_id: the ID of the related comment at the exporting component
    :param component_id: the ID of the exporting component
    :return: the ID of the new comment
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """

    assert (component_id is None) == (fed_id is None)
    assert component_id is not None or user_id is not None

    # ensure that the object exists
    objects.check_object_exists(object_id)
    if user_id is not None:
        # ensure that the user exists
        users.check_user_exists(user_id)
    if component_id is not None:
        # ensure that the component can be found
        components.check_component_exists(component_id)
    comment = models.Comment(
        object_id=object_id,
        user_id=user_id,
        content=content,
        fed_id=fed_id,
        component_id=component_id,
        utc_datetime=utc_datetime
    )
    db.session.add(comment)
    db.session.commit()
    if component_id is None and create_log_entry:
        # ensured by the if at the start of the function
        assert user_id is not None
        object_log.post_comment(user_id=user_id, object_id=object_id, comment_id=comment.id)
        user_log.post_comment(user_id=user_id, object_id=object_id, comment_id=comment.id)
    comment_id: int = comment.id
    return comment_id


def get_comment(comment_id: int, component_id: typing.Optional[int] = None) -> Comment:
    """
    :param comment_id: the federated ID of the comment
    :param component_id: the components ID (source)
    :return: the comment
    :raise errors.CommentDoesNotExistError: when no comment with the given ID exists
    """
    if component_id is None:
        comment = models.Comment.query.filter_by(id=comment_id).first()
    else:
        # ensure that the component can be found
        components.check_component_exists(component_id)
        comment = models.Comment.query.filter_by(fed_id=comment_id, component_id=component_id).first()
    if comment is None:
        raise errors.CommentDoesNotExistError()
    return Comment.from_database(comment)


def update_comment(
        comment_id: int,
        user_id: typing.Optional[int],
        content: str,
        utc_datetime: datetime.datetime
) -> None:
    """
    :param comment_id: the ID of an existing comment
    :param user_id: the ID of the (new) author
    :param content: the new content
    :param utc_datetime: the new datetime
    :raise errors.CommentDoesNotExistError: when no comment with the given ID
        exists
    """
    comment = models.Comment.query.filter_by(id=comment_id).first()
    if comment is None:
        raise errors.CommentDoesNotExistError()
    # user_id must not be None for local comments
    assert comment.component_id is not None or user_id is not None
    comment.user_id = user_id
    comment.content = content
    comment.utc_datetime = utc_datetime
    db.session.add(comment)
    db.session.commit()


def get_comments_for_object(object_id: int) -> typing.List[Comment]:
    """
    Returns a list of comments for an object.

    :param object_id: the ID of an existing object
    :return: the list of comments, sorted from oldest to newest
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    comments = models.Comment.query.filter_by(object_id=object_id).order_by(db.asc(models.Comment.utc_datetime)).all()
    if not comments:
        # ensure that the object exists
        objects.check_object_exists(object_id)
    return [
        Comment.from_database(comment)
        for comment in comments
    ]


def get_comment_for_object(object_id: int, comment_id: int) -> Comment:
    """
    Returns a specific comment for a given object.

    :param object_id: the ID of an existing object
    :param comment_id: the ID of an existing comment
    :return: the specified comment
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.CommentDoesNotExistError: when no comment with the given
        comment ID exists for this object
    """
    comment = models.Comment.query.filter_by(object_id=object_id, id=comment_id).first()
    if not comment:
        # ensure that the object exists
        objects.check_object_exists(object_id)
        raise errors.CommentDoesNotExistError()
    return Comment.from_database(comment)



# coding: utf-8
"""

"""

import collections
import dataclasses
import typing

from .. import db
from .. import models
from . import errors
from . import object_log
from . import user_log
from . import object_permissions
from .objects import check_object_exists
from .users import check_user_exists


@dataclasses.dataclass(frozen=True)
class Publication:
    """
    This class provides an immutable wrapper around models.object_publications.ObjectPublication.
    """
    doi: str
    title: typing.Optional[str]
    object_name: typing.Optional[str]

    @classmethod
    def from_database(cls, publication: models.object_publications.ObjectPublication) -> 'Publication':
        return Publication(
            doi=publication.doi,
            title=publication.title,
            object_name=publication.object_name
        )


def link_publication_to_object(
        user_id: int,
        object_id: int,
        doi: str,
        title: typing.Optional[str] = None,
        object_name: typing.Optional[str] = None
) -> None:
    """
    Link a publication to an object.

    :param user_id: the ID of an existing user
    :param object_id: the ID of an existing object
    :param doi: the DOI of an existing publication
    :param title: the title of the publication
    :param object_name: the name of the object in the publication
    :raise errors.ObjectDoesNotExistError: when no object with the given object
        ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    """
    check_user_exists(user_id)
    check_object_exists(object_id)
    link = models.object_publications.ObjectPublication.query.filter_by(object_id=object_id, doi=doi).first()
    if link is None:
        link = models.object_publications.ObjectPublication(object_id=object_id, doi=doi)
    link.title = title
    link.object_name = object_name
    db.session.add(link)
    db.session.commit()
    user_log.link_publication(user_id=user_id, object_id=object_id, doi=doi, title=title, object_name=object_name)
    object_log.link_publication(user_id=user_id, object_id=object_id, doi=doi, title=title, object_name=object_name)


def get_publications_for_object(object_id: int) -> typing.Sequence[Publication]:
    """
    Get a list of publications linked to an object with a given ID.

    :param object_id: the ID of an existing object
    :return: a list of publications linked to the given object
    :raise errors.ObjectDoesNotExistError: when no object with the given object ID
        exists
    """
    check_object_exists(object_id)
    links = models.object_publications.ObjectPublication.query.filter_by(object_id=object_id).all()
    return [
        Publication.from_database(link)
        for link in links
    ]


def get_object_ids_linked_to_doi(doi: str) -> typing.Sequence[int]:
    """
    Get a list of object IDs linked to a publication.

    :param doi: the DOI of the publication
    :return: the list of object IDs
    :raise errors.InvalidDOIError: when the DOI is invalid
    """
    doi = simplify_doi(doi)
    links = models.object_publications.ObjectPublication.query.filter_by(doi=doi).all()
    return [
        link.object_id
        for link in links
    ]


def simplify_doi(doi: str) -> str:
    """
    Simplify a digital object identifier (DOI).

    :param doi: the DOI to simplify
    :return: the simplified DOI
    :raise errors.InvalidDOIError: when the DOI is invalid
    """
    doi = doi.lower()
    doi_prefixes = [
        'doi:',
        'https://doi.org/'
    ]
    for doi_prefix in doi_prefixes:
        if doi.startswith(doi_prefix):
            doi = doi[len(doi_prefix):]
            break
    # DOI organizer part must start with '10.'
    if not doi.startswith('10.'):
        raise errors.InvalidDOIError()
    # DOI must be split into organizer and object parts by a slash
    if '/' not in doi[4:-1]:
        raise errors.InvalidDOIError()
    organizer_id, object_id = doi[3:].split('/', 1)
    if any(c not in '0123456789.' for c in organizer_id):
        raise errors.InvalidDOIError()
    if any(c in '<>"' for c in object_id):
        raise errors.InvalidDOIError()
    return doi


def get_publications_for_user(user_id: int) -> typing.List[typing.Tuple[str, typing.Optional[str]]]:
    """
    Get a list of DOIs and titles for publications linked to objects readable by the given user.

    If multiple titles are used for the same DOI, the most used, non-empty
    title will be used, with lexicographical sorting as final criteria.

    :param user_id: the ID of an existing user
    :return: a list containing publication DOIs and titles
    """
    table, parameters = object_permissions.get_object_table_with_permissions(
        user_id=user_id,
        permissions=models.Permissions.READ,
        name_only=True
    )
    if table is None:
        return []
    publication_links = [
        row[0] for row in db.session.execute(
            db.select(models.ObjectPublication).select_from(
                table.join(
                    models.ObjectPublication,
                    table.c.object_id == models.ObjectPublication.object_id
                )
            ), parameters
        ).fetchall()
    ]
    publication_titles: typing.Dict[str, typing.List[str]] = {}
    for link in publication_links:
        if link.doi not in publication_titles:
            publication_titles[link.doi] = []
        if link.title:
            publication_titles[link.doi].append(link.title)
    publications = []
    for doi, titles in publication_titles.items():
        title: typing.Optional[str]
        if titles:
            # sort titles lexicographically, then find most commonly used title
            titles.sort()
            title = collections.Counter(titles).most_common(1)[0][0]
        else:
            title = None
        publications.append((doi, title))
    publications.sort()
    return publications



# coding: utf-8
"""
Add fed_object_id column to objects_previous table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('objects_previous', 'fed_object_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE objects_previous
            ADD fed_object_id INTEGER,
            ADD fed_version_id INTEGER,
            ADD component_id INTEGER,
            ADD FOREIGN KEY(component_id) REFERENCES components(id)
    """))
    return True
